/*
 * A queue mechanism for concurrent access to a resource.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/util/cap_alloc.h>

#include <stdlib.h>

#include <pthread-l4.h>

#include <ipc/server.h>
#include <ipc/util_ipc.h>
#include "queue.h"

/* Client and server interfaces. */

#include "queue_client.h"
#include "queue_server.h"



/* Thread main function for a queue. */

static void *queue_thread(void *data)
{
  QueueObject *obj = reinterpret_cast<QueueObject *>(data);

  obj->mainloop();
  return 0;
}



/* Return a queue client. */

Queue *QueueObject::client()
{
  if (!_client)
    _client = new client_Queue(_thread_ep);
  return _client;
}

/* Set up thread. */

long QueueObject::start()
{
  pthread_create(&_thread, NULL, queue_thread, this);

  long err = ipc_server_new_gate_for_thread(&_thread_ep, pthread_l4_cap(_thread), l4_umword_t(this));
  if (err)
    return err;

  return L4_EOK;
}

/* Join to the thread. */

long QueueObject::end()
{
  void *retval;

  pthread_join(&_thread, &retval);
  return (long) retval;
}

/* Listen for messages. */

void QueueObject::mainloop()
{
  l4_umword_t label;

  ipc_message_expect(&_msg, Queue_expected_items);

  while (1)
  {
    ipc_message_wait(&_msg, &label);

    /* Ignore erroneous messages. */

    if (l4_ipc_error(_msg.tag, l4_utcb()))
      continue;

    /* Open and dispatch the message to the appropriate method.
       Replies are sent implicitly only for operations which do not need a
       separate completion step, such as the leave method. */

    handle_Queue(&_msg, this);

    if (_msg.terminating)
      break;
  }
}

/* Remove the running task from the queue, releasing its capability. */

void QueueObject::remove()
{
  if (l4_is_invalid_cap(_running))
    return;

  l4re_util_cap_free_um(_running);

  _running = L4_INVALID_CAP;
}

/* Schedule another user from the queue. */

void QueueObject::next()
{
  if (_queue.empty())
    return;

  if (l4_is_valid_cap(_running))
    _queue.push_back(_running);

  _running = pop();

  /* Yield to the explicitly nominated initiator. */

  complete_Queue_yield(_running);
}

/* Pop the front of the queue. */

l4_cap_idx_t QueueObject::pop()
{
  l4_cap_idx_t front = _queue.front();
  _queue.pop_front();
  return front;
}

/* Handle a new queue member. */

void QueueObject::new_member()
{
  /* Only schedule another user if another is not running and is
     available. */

  if (_queue.size() > 1)
    if (l4_is_invalid_cap(_running))
      next();
}

/* Register the user and wait to continue. */

long QueueObject::join(l4_cap_idx_t user)
{
  _queue.push_back(user);
  new_member();
  return L4_EOK;
}

/* Register the user and either continue or wait to continue. */

long QueueObject::join_first(l4_cap_idx_t user)
{
  _queue.push_front(user);
  new_member();
  return L4_EOK;
}

/* Remove the user and schedule another user. */

long QueueObject::leave()
{
  /* A reply is sent in the wrapper for this operation. */

  remove();
  next();

  return L4_EOK;
}

/* Yield from the user, scheduling another. */

long QueueObject::yield()
{
  next();

  return L4_EOK;
}
